#include <stdio.h>
#include <conio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include "watdefs.h"
#include "afuncs.h"

#define NAMESORT_RECLEN 19

/* Small utility to process the WDS data from USNO to ensure that
   an RA/dec field,  of either high or low precision,  is present
   in the final columns.  It also removes some of the extraneous
   HTML tags and blank lines,  and sorts the output in RA order.
   (The input is almost,  but not quite,  in RA order.)  It also creates
   the 'namesort.dat' file needed for Guide's "Go To... Double Star"
   function.  */

int name_compare( const void *str1, const void *str2)
{
   return( memcmp( str1, str2, NAMESORT_RECLEN));
}

   /* In sorting the WDS,  just look at the text in the RA/dec field: */

int wds_compare( const void *str1, const void *str2)
{
   return( memcmp( (const char *)str1 + 112,
                   (const char *)str2 + 112, 18));
}

static double grab_angle_from_buff( const char *iptr)
{
   const long factors[6] = { 36000, 3600, 600, 60, 10, 1 };
   long rval = 0;
   double drval;
   char tbuff[6];
   int i;

   for( i = 0; i < 6; i++)
      if( iptr[i] > '0' && iptr[i] <= '9')
         rval += (long)( iptr[i] - '0') * factors[i];
   drval = (double)rval;
   memcpy( tbuff, iptr + 6, 5);
   tbuff[5] = '\0';
   if( tbuff[0] == '.')
      drval += atof( tbuff);
   return( drval / 3600.);    /* cvt to degrees or decimal hours */
}

static void put_ra_dec_into_name_data( const char *ra_dec_in,
                               char *ra_dec_out)
{
   const double pi = 3.14159265358979323846;
   DPT ipt, opt;
   long tval;

   ipt.x = -grab_angle_from_buff( ra_dec_in) * (pi / 12.);
   ipt.y = grab_angle_from_buff( ra_dec_in + 10) * (pi / 180.);
   if( ra_dec_in[9] == '-')
      ipt.y *= -1.;
   precess_pt( &opt, &ipt, 2000., 1900.);
   opt.x = fmod( opt.x, 2. * pi);
   while( opt.x < 0.)
      opt.x += pi * 2.;
   while( opt.x > pi * 2.)
      opt.x -= pi * 2.;

   tval = 24 * 600 - (long)( opt.x * (12. / pi) * 600. + .5);
   sprintf( ra_dec_out, "%02ld%03ld", tval / 600L, tval % 600L);
   if( opt.y < 0.)
      {
      ra_dec_out[5] = '-';
      opt.y = -opt.y;
      }
   else
      ra_dec_out[5] = '+';
   tval = (long)floor( opt.y * (180. / pi) * 60. + .5);
   sprintf( ra_dec_out + 6, "%02ld%02ld", tval / 60L, tval % 60L);
}

int main( int argc, char **argv)
{
   FILE *ifile = fopen( "wdsweb_summ.txt", "rb"), *ofile;
   char *namesort_data, *irecs;
   char tbuff[300];
   size_t record_length, n_records, i;
   int percent_done = 0;

   setvbuf( stdout, NULL, _IONBF, 0);
   if( !ifile)
      ifile = fopen( "wdsweb~2.txt", "rb");
   if( !ifile)
      {
      printf( "'wdsweb_summ.txt' wasn't found.  Check to be sure that it's in the\n");
      printf( "Guide directory.  If it's not there,  move it there and try\n");
      printf( "running 'add_loc' again.\n");
      printf( "Hit any key:\n");
      getch( );
      return( -1);
      }

   fgets( tbuff, sizeof( tbuff), ifile);
   record_length = strlen( tbuff);
   fseek( ifile, 0L, SEEK_END);
   n_records = ftell( ifile) / record_length;
   printf( "Reading %d records of %d bytes each\n", n_records, record_length);
   namesort_data = (char *)malloc( n_records * NAMESORT_RECLEN);
   irecs = (char *)malloc( n_records * record_length);
   fseek( ifile, 0L, SEEK_SET);
   fread( irecs, record_length, n_records, ifile);
   fclose( ifile);
   printf( "Processing records...\n");
   for( i = 0; i < n_records; i++)
      {
      char *buff = irecs + i * record_length;
      char *nameptr = namesort_data + i * NAMESORT_RECLEN;
      int new_pct;

      if( buff[117] == ' ')     /* no precise RA/dec given,  so... */
         {
         memcpy( buff + 112, buff, 4); /* copy in HHMM from RA */
         sprintf( buff + 116, "%02d   ", (buff[4] - '0') * 6);
         memcpy( buff + 121, buff + 5, 5);  /* copy in +/-ddmm */
         memcpy( buff + 126, "   ", 3);
         }
      memcpy( nameptr, buff + 10, 8);     /* copy object name... */
      put_ra_dec_into_name_data( buff + 112, nameptr + 8);  /* and put in loc */
      nameptr[NAMESORT_RECLEN - 1] = (char)10;
      new_pct = (int)( i * 100 / n_records);
      if( new_pct != percent_done)
         {
         const double t_done = (double)clock( ) / CLOCKS_PER_SEC;
         const double t_remain =
           t_done * (double)( n_records - i) / (double)i;

         percent_done = new_pct;
         printf( "%d%% done; %.0lf seconds elapsed, %.0lf seconds remain\r",
                    percent_done, t_done, t_remain);
         }
      }
   printf( "\nSorting WDS records in RA order...\n");
   qsort( irecs, n_records, record_length, wds_compare);
   printf( "\nWriting WDS data to 'wds_out.txt'...\n");
   ofile = fopen( "wds_out.txt", "wb");
   fwrite( irecs, record_length, n_records, ofile);
   fclose( ofile);
   free( irecs);
   printf( "\nWDS data file successfully converted\n");

   ofile = fopen( "namesort.dat", "wb");
   qsort( namesort_data, n_records, NAMESORT_RECLEN, name_compare);
   fwrite( namesort_data, n_records, NAMESORT_RECLEN, ofile);
   fclose( ofile);
   if( argc > 1)        /* output list of discoverers */
      for( i = 1; i < n_records; i++)
         if( memcmp( namesort_data +   i   * NAMESORT_RECLEN,
                     namesort_data + (i-1) * NAMESORT_RECLEN, 3))
            {
            memcpy( tbuff, namesort_data + i * NAMESORT_RECLEN, 3);
            tbuff[3] = '\0';
            printf( "%s\n", tbuff);
            }
   printf( "Hit any key:\n");
   getch( );
   return( 0);
}
